"""
    Created on 2012-01-19
    @author: jldupont
"""
from   functools import partial
from   time import sleep
import logging, sys, os, json
import boto

from boto.sqs.jsonmessage import JSONMessage
from boto.sqs.message     import RawMessage

from tools_sys import retry

def stdout(s, json_mode=False):
    if json_mode:
        s=json.dumps(s)
    try:
        if not s.endswith("\n"):
            s=s+"\n"
    except:
        pass
    sys.stdout.write(s)

def run(queue_name=None, 
        queue_name_output=None,
        write_delay=None,
        flush_queue=None,
        batch_size=None,
        format_any=None,
        trigger_none_msg=None, trigger_topic=None,
        delete_on_error=False, dont_pass_through=False,
        simulate_error=False, error_msg=None,
        log_network_error=False,
        loglevel=None ):
    
    if write_delay>900 or write_delay<1:
        raise Exception("Invalid value for 'delay_seconds': must be between 1 and 900")
    
    # SETUP PRIVATE QUEUE
    #####################
    def setup_queue(qn):
        """ closure for setting up private queue """
        conn = boto.connect_sqs()
        q=conn.create_queue(qn)
        if format_any:
            q.set_message_class(RawMessage)
        else:
            q.set_message_class(JSONMessage)
        return q 
        
    logging.info("Creating Input Queue... (automatic retry)")
    try:
        qi=retry(partial(setup_queue, queue_name))
        logging.info("Input Queue successfully created") 
    except Exception,e:
        raise Exception("Creating queue '%s': %s" % (queue_name, str(e)))

    logging.info("Creating Output Queue... (automatic retry)")
    try:
        qo=retry(partial(setup_queue, queue_name_output))
        logging.info("Output Queue successfully created") 
    except Exception,e:
        raise Exception("Creating queue '%s': %s" % (queue_name_output, str(e)))

    # FLUSH QUEUE
    #############
    if flush_queue:
        try:    
            qi.clear()
            logging.info("Input Queue flushed")
        except: pass


    # MAIN LOOP
    ###########
    ppid=os.getppid()
    logging.info("Process pid: %s" % os.getpid())
    logging.info("Parent pid:  %s" % ppid)
    logging.info("Starting loop...")
    while True:
        if os.getppid()!=ppid:
            logging.warning("Parent terminated... exiting")
            break
        
        line=sys.stdin.readline().strip()

        if len(line)==0:
            continue

        logging.debug("Received '%s' from stdin" % line)

        if not dont_pass_through:
            stdout(line)
        
        try:
            jo=json.loads(line)
        except:
            logging.error("Can't JSON decode: %s" % line)
            continue
        
        try:
            topic=jo["topic"]
            if topic!=trigger_topic:
                logging.debug("Message trigger not found (topic was: %s)" % topic)
                continue
        except:
            logging.error("Input JSON doesn't have a 'topic' field")
            continue
                   
        try:
            if simulate_error:
                raise Exception("Network error simulation")
            
            msgs=qi.get_messages(num_messages=batch_size)
        except Exception, e:
            if log_network_error:
                logging.error("Network error - rx input queue")
            
            if error_msg is not None:
                try:    msg=error_msg % str(e)
                except: msg=error_msg
                stdout(msg)
            continue
                
        
        ### take care of sending a specified string when
        ### there are no messages in the queue
        if msgs is None or len(msgs)==0:
            if trigger_none_msg is not None:
                stdout(trigger_none_msg)
            continue
            
        ### normal flow
        for msg in msgs:
            
            try:
                b=msg.get_body()
            except:
                logging.error("Can't retrieve msg body - continue to next msg")
                if delete_on_error:
                    qi.delete_message(msg)                
                continue
            
            logging.debug("Got 1 message to process...")
            
            try:
                stdout(b, json_mode=not format_any)
            except:
                ### exit because of probable broken pipe
                raise Exception("Can't write to stdout (probably broken pipe)")
            
            ## write to output queue
            ## don't fail too fast :)
            logging.debug("Attempting to enqueue in output queue...")
            try:
                if not qo.write(msg, delay_seconds=write_delay):
                    raise
            except:
                logging.debug("Error writing to output queue -- sleeping a bit before retrying")
                sleep(1)
                try:
                    if not qo.write(msg, delay_seconds=write_delay):
                        raise
                except:
                    logging.debug("Error writing to output queue -- giving up")
                    if log_network_error:
                        logging.error("Can't write message in output queue - probable job duplication to ensue")
                continue
                
            ## After we have successfully queued the message for the output queue,
            ## delete the source message
            logging.debug("Deleting source message from input queue...")
            try:
                qi.delete_message(msg)
            except Exception, e:
                logging.debug("Error deleting source message -- sleeping a bit before retrying")
                sleep(1)
                try:
                    qi.delete_message(msg)
                except:
                    logging.debug("Error deleting source message -- giving up")
                    if log_network_error:
                        logging.error("Can't delete received msg - probable job duplication to ensue")
        





"""
    Created on 2012-01-20
    @author: jldupont
"""
import logging
import types

def pprint_kv(k,v, align=20):
    fmt="%-"+str(align)+"s : %s"
    print fmt % (k, v)

def info_dump(d, align):
    fmt="%-"+str(align)+"s : %s"

    if type(d)==types.DictionaryType:        
        for key in d:
            logging.info(fmt % (key, d[key]))
            
    if type(d)==types.ListType:
        for el in d:
            key, value=el
            logging.info(fmt % (key, value))
        
def setloglevel(level_name):
    """
    >>> import logging
    >>> setloglevel("info")
    >>> logging.debug("test")
    
    """
    try:
        ll=getattr(logging, level_name.upper())
        logger=logging.getLogger()
        logger.setLevel(ll)
    except:
        raise Exception("Invalid log level name: %s" % level_name)



"""
    @author: Jean-Lou Dupont
"""

import logging,sys,os

## force stdout + stderr flushing
sys.stdout = os.fdopen(sys.stdout.fileno(), 'w', 0)
sys.stderr = os.fdopen(sys.stderr.fileno(), 'w', 0)

name=os.path.basename(sys.argv[0])
fname="%-10s" % name

FORMAT='%(asctime)s - '+fname+' - %(levelname)s - %(message)s'
formatter = logging.Formatter(FORMAT)

logging.basicConfig(level=logging.INFO, format=FORMAT)



"""
    Created on 2012-01-19
    @author: jldupont
"""
import time
from datetime import datetime, timedelta


def get_date():
    """
    Returns a string of the format "YYYYMMDD"
    """
    now=datetime.now()
    s="%s%s%s" % (now.year, str(now.month).zfill(2), str(now.day).zfill(2))
    return (now, s)


def get_timestamp():
    """
    Returns a string of the format "YYYYMMDDHHMMSS"
    """
    now, s=get_date()
    return (now, "%s%s%s%s" % (s, str(now.hour).zfill(2), str(now.minute).zfill(2), str(now.second).zfill(2)))


def iso_now():
    """
    YYYY-MM-DDTHH:MM:SS.xxxxxx
    
    e.g. '2012-01-25T08:58:27.514373'
    """
    now=datetime.today()
    return (now, now.isoformat())


def utc_iso_now():
    """
    Returns the present time @ GMT UTC
    
    time.struct_time(tm_year=2011, tm_mon=6, tm_mday=28, tm_hour=16, tm_min=22, tm_sec=55, tm_wday=1, tm_yday=179, tm_isdst=0)
    
    @return (time.struct_time, string)
    """
    now=time.gmtime()
    fmt="%Y-%m-%dT%H:%M:%S"
    r=time.strftime(fmt, now)
    return (now, r)

def iso_from_string(inp):
    """
    need to remove the trailing miliseconds
    
    >>> iso_from_string("2012-01-25T08:58:27.514373")
    datetime.datetime(2012, 1, 25, 8, 58, 27)
    """
    try:     r=inp.split(".")[0]
    except:  r=inp
    fmt="%Y-%m-%dT%H:%M:%S"
    return datetime.strptime(r, fmt)

def iso_string_add_delta(inp, days=0, hours=0, minutes=0):
    """
    Adds the specified amount of time to the input iso date
    >>> now, nowstr=utc_iso_now()
    >>> t, tstr=iso_string_add_delta(nowstr, days=1, hours=12)
    >>> print t.day-now.tm_mday
    1
    >>> print t.hour-now.tm_hour
    12
    """
    dt=iso_from_string(inp)
    td=timedelta(days=days, hours=hours, minutes=minutes)
    r=dt+td
    return (r, r.isoformat())


def iso_now_and_then(days=0, hours=0, minutes=0):
    """
    return e.g. ('2012-01-25T13:58:27', '2012-01-27T01:58:27')
    
    >>> now, then=iso_now_and_then(days=1, hours=12)
    >>> r=iso_delta_from_strings(now, then)
    >>> print r.days
    1
    >>> print r.seconds
    43200
    """
    _, nowstr=utc_iso_now()
    _, thenstr=iso_string_add_delta(nowstr, days=days, hours=hours, minutes=minutes)
    return (nowstr, thenstr)
    


def iso_delta_from_strings(t1, t2):
    """
    Computes the delta (in seconds) between t1 and t2
    
    >>> import time
    >>> _, t1=utc_iso_now()
    >>> time.sleep(2)
    >>> _, t2=utc_iso_now()
    >>> delta=iso_delta_from_strings(t1, t2)
    >>> print delta.seconds
    2
    >>> print delta.days
    0
    
    @return datetime.timedelta
    """
    t1c=iso_from_string(t1)
    t2c=iso_from_string(t2)
    return t2c-t1c


if __name__=="__main__":
    import doctest
    doctest.testmod()


"""
    Created on 2012-02-08
    @author: jldupont
"""
from pyfnc import patterned, pattern, coroutine


@pattern(None, None)
def is_trans_NN(previous, current):
    return ('nop', None)

@pattern(None, bool)
def is_trans_NB(previous, current):
    return ('tr', "up" if current is True else "down")
    
@pattern(False, False)
def is_trans_FF(previous, current):
    return ("nop", None)

@pattern(True, True)
def is_trans_TT(previous, current):
    return ("nop", None)
    
@pattern(True, False)
def is_trans_TF(previous, current):
    return ("tr", "down")

@pattern(False, True)
def is_trans_FT(previous, current):
    return ("tr", "up")
    
@patterned
def is_trans(previous, current):
    """
    >>> is_trans(True, True)
    ('nop', None)
    >>> is_trans(False, False)
    ('nop', None)
    >>> is_trans(True, False)
    ('tr', 'down')
    >>> is_trans(False, True)
    ('tr', 'up')
    >>> is_trans(None, None)
    ('nop', None)
    >>> is_trans(None, True)
    ('tr', 'up')
    >>> is_trans(None, False)
    ('tr', 'down')
    """

@coroutine
def transition_manager(ctx):
    """
    ctx[param]={ "state": state, "up": fnc, "down": fnc}
    
    >>> import logging
    >>> fu=lambda:logging.warning("up")
    >>> fd=lambda:logging.warning("down")
    >>> ctx={"test":{"up": fu, "down":fd}}
    >>> tm=transition_manager(ctx)
    >>> tm.send(("test", True))   ### warning output 'up'
    >>> tm.send(("test", True))
    >>> tm.send(("test", False))  ### warning output 'down'
    >>> tm.send(("test", False))
    """    
    while True:
        param, current_state=(yield)
        cparam=ctx.get(param, {})
        
        previous_state=cparam.get("previous", None)
        maybe_tr, maybe_dir=is_trans(previous_state, current_state)
        cparam["previous"]=current_state
        ctx[param]=cparam
        
        if maybe_tr=="tr":
            fnc=cparam.get(maybe_dir, None)
            if fnc is not None:
                fnc()
        
        
if __name__=="__main__":
    import doctest
    doctest.testmod()



"""
    Created on 2012-01-19
    @author: jldupont
"""
import logging, os, json, importlib
from time import sleep
import boto
import json
import sys

from boto.exception import SQSDecodeError
from boto.sqs.jsonmessage import JSONMessage

from tools_sqs import gen_queue_name
from tools_sns import build_topic_arn
from tools_mod import call

def stdout(s):
    sys.stdout.write(s+"\n")

def process_msg(args, queue_name, topic, module_name, msg):
    
    result=call(module_name, "run", queue_name, topic, msg.get_body())
    if result is not None:
        logging.info("result: %s" % str(result))
        
def maybe_json_to_stdout(enable_json, queue_name, topic, msg):
    if enable_json:
        
        try:
            mo=msg.get_body()
            
            o={ "queue_name": queue_name
              ,"topic": topic
              ,"msg": mo
              }
            stdout(json.dumps(o))
        except Exception, e:
            logging.error("can't output json to stdout: %s" % str(e))

def run(args):
    """
    0. check 'module_name'
    1. create private SQS queue
    2. subscribe queue to the specified 'topic'
    """
    enable_json=args.enable_json
    module_name=args.module_name
    enable_call_run=True if module_name.lower()!="none" else False
    batch_size=args.batch_size
   
    logging.info("module_name=      %s" % module_name)    
    logging.info("batch_size=       %s" % batch_size)
    logging.info("polling_interval= %s (seconds)" % args.polling_interval)
    logging.info("json to stdout=   %s" % enable_json)
        
    # SETUP PRIVATE QUEUE
    try:
        conn = boto.connect_sqs()  
        queue_name=gen_queue_name()
        q=conn.create_queue(queue_name)
        q.set_message_class(JSONMessage) 
        
    except Exception,e:
        raise Exception("Creating queue '%s': %s" % (queue_name, str(e)))

    logging.info("queue_name=       %s" % queue_name)
    
    topic=args.topic
    topic_arn=build_topic_arn(conn, topic)
    
    logging.info("topic=            %s" % topic)
    logging.info("topic_arn=        %s" % topic_arn)    
    
    # SUBSCRIBE TO TOPIC
    try:
        snsconn=boto.connect_sns()
        snsconn.subscribe_sqs_queue(topic_arn, q)
        logging.info("Subscribed to topic '%s'" % topic)
    except Exception,e:
        raise Exception("Subscribing to topic '%s': %s" % (topic, str(e)))
    
    ppid=os.getppid()
    logging.info("Process pid: %s" % os.getpid())
    logging.info("Parent pid: %s" % ppid)
    logging.info("Starting loop...")
    while True:
        if os.getppid()!=ppid:
            logging.warning("Parent terminated... exiting")
            break
        try:
            msgs=q.get_messages(num_messages=batch_size)
            if msgs is not None:
                for msg in msgs:
                    if enable_call_run:
                        process_msg(args, queue_name, topic, module_name, msg)
                    maybe_json_to_stdout(enable_json, queue_name, topic, msg)
                    q.delete_message(msg)
                
        except SQSDecodeError:
            logging.warning("Message decoding error")
            
        logging.debug("...sleeping for %s seconds" % args.polling_interval)
        sleep(args.polling_interval)
        



"""
    Created on 2012-01-19
    @author: jldupont
"""
import logging, sys, os, json
import boto
from time import sleep

from boto.sqs.jsonmessage import JSONMessage
from boto.sqs.message     import RawMessage

from tools_sys import retry

MAX_ERROR_COUNT=25

def stdout(s):
    try:
        if not s.endswith("\n"):
            s=s+"\n"
    except:
        pass
    sys.stdout.write(s)
    sys.stdout.flush()

def run(queue_name=None, flush_queue=None,
        batch_size=None, polling_interval=None,
        format_any=None, propagate_error=None,
        retry_always=None, wait_trigger=None,
        trigger_none_msg=None, trigger_topic=None,
        delete_on_error=False, dont_pass_through=False,
        simulate_error=False, error_msg=None ):
    
    ## we need a minimum of second between polls
    polling_interval=max(1, polling_interval)
   
    # SETUP PRIVATE QUEUE
    #####################
    def setup_private_queue():
        """ closure for setting up private queue """
        conn = boto.connect_sqs()
        q=conn.create_queue(queue_name)
        if format_any:
            q.set_message_class(RawMessage)
        else:
            q.set_message_class(JSONMessage)
        return q 
        
    try:
        q=retry(setup_private_queue, always=retry_always)
        logging.info("Queue successfully created") 
    except Exception,e:
        raise Exception("Creating queue '%s': %s" % (queue_name, str(e)))

    # FLUSH QUEUE
    #############
    if flush_queue:
        try:    
            q.clear()
            logging.info("Queue flushed")
        except: pass

    error_count=0

    # MAIN LOOP
    ###########
    ppid=os.getppid()
    logging.info("Process pid: %s" % os.getpid())
    logging.info("Parent pid: %s" % ppid)
    logging.info("Starting loop...")
    while True:
        if os.getppid()!=ppid:
            logging.warning("Parent terminated... exiting")
            break
        
        
        if wait_trigger or trigger_topic is not None:
            line=sys.stdin.readline().strip()

            if not dont_pass_through:
                stdout(line)

        if trigger_topic is not None:
            if len(line)==0:
                continue
            try:
                jo=json.loads(line)
            except:
                logging.error("Can't JSON decode: %s" % line)
                continue
            
            try:
                topic=jo["topic"]
                #logging.info("Found topic: %s" % topic)
                if topic!=trigger_topic:
                    continue
            except:
                logging.error("Input JSON doesn't have a 'topic' field")
                continue
            
       
        try:
            if simulate_error:
                raise Exception("Network error simulation")
            
            msgs=q.get_messages(num_messages=batch_size)
            error_count=0
        except Exception, e:
            
            if error_msg is not None:
                try:    msg=error_msg % str(e)
                except: msg=error_msg
                stdout(msg)
                continue
                
            if propagate_error:
                stdout('''{"error": "%s"}''' % str(e))
                
            if retry_always:
                continue
        
            error_count=error_count+1
            if error_count<MAX_ERROR_COUNT:
                continue
            
            raise Exception("Exiting because of excessive network or decode error")
        
        ### take care of sending a specified string when
        ### there are no messages in the queue & we are in "trigger mode"
        if msgs is None or len(msgs)==0:
            if wait_trigger or trigger_topic is not None:
                if trigger_none_msg is not None:
                    stdout(trigger_none_msg+"\n")
                    continue
            
        ### normal flow
        if msgs is not None:
            for msg in msgs:
                
                try:
                    b=msg.get_body()
                                        
                    if format_any:
                        stdout(b+"\n")
                    else:
                        stdout(json.dumps(b)+"\n")
                        
                    q.delete_message(msg)
                except Exception, e:
                    if delete_on_error:
                        q.delete_message(msg)
                    logging.error("Can't process received msg: %s --> %s" % (str(b), e))
        
        if not wait_trigger and trigger_topic is None:
            logging.debug("...sleeping for %s seconds" % polling_interval)
            sleep(polling_interval)

